package com.taco.debug.gui;

import java.lang.reflect.*;
import java.beans.*;
import java.io.Serializable;
import java.util.Collection;
import java.util.Collections;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;

import java.awt.Image;

import javax.swing.Icon;
import javax.swing.ImageIcon;

import javax.swing.tree.TreeNode;

import org.apache.log4j.Logger;

import com.taco.lang.reflect.ReflectionUtilities;

/** An implementation of <code>TreeNode</code> that represents a bean. */
public class BeanTreeNode implements TreeNode, Serializable {
    /** Create a root node with the argument name. */ 
    public BeanTreeNode(String name, Object obj) {
        this(null, name, obj, true);
    }
    
    public BeanTreeNode(TreeNode parent, String name, Object obj, 
      boolean shouldIntrospect) {
        _name = name;
        _parent = parent;
        
        _obj = obj;
        _shouldIntrospect = shouldIntrospect;      
    }
    
    public Enumeration children() {
        _initialize();
        return Collections.enumeration(_children);
    }
           
    public boolean getAllowsChildren() {
        _initialize();
        return _allowsChildren;   
    }
           
    public TreeNode getChildAt(int childIndex) {
        _initialize();
        return (TreeNode) _children.get(childIndex);
    }

         
    public int getChildCount() {
        _initialize();
        return _children.size();
    }
 
    public int getIndex(TreeNode node) {
        _initialize();
        for (int i = 0; i < _children.size(); i++) {
            if (_children.get(i).equals(node)) {
                return i;                
            }            
        }
        
        // Reached end without finding the node.
        return -1;
    }
     
    public TreeNode getParent() { return _parent; }
           
    public boolean isLeaf() {
        _initialize();
        return _isLeaf;
    }

    /** Return the name of this node. */
    public String getName() {
        return _name;
    }
 
    /** Return the object that this node describes. */
    public Object getObject() {
        return _obj;
    }
    
    public Icon getIcon() {
        _initialize();
        return _icon;
    }              
    
    public String toString() {
        return getName() + " : " + _obj;
    }
        
    /** Return the maximum number of array or Collection elements. By default,
     *  this returns 4.
     */
    public int getMaxLength() {
        return 4;   
    }
    
    protected void _initialize() {
        if (_initialized || !_shouldIntrospect) {
            return;   
        }        
         
        if (_logger.isDebugEnabled()) { 
            _logger.debug("initialize " + _obj);
        }
        
        try {
            if (_obj == null) {
                _allowsChildren = false;
                _isLeaf = true;
                return;
            } 
            
            Class clazz = _obj.getClass();
            
            if (ReflectionUtilities.isPrimitiveWrapperClass(clazz)) {
                _allowsChildren = false;
                _isLeaf = true;
            } else if (clazz.isArray()) {                
                _initializeFromArray();                
            } else {
                _allowsChildren = true;
                _isLeaf = false;
                
                // If the object is an instance of collection, 
                // add its elements.                                                
                if (Collection.class.isAssignableFrom(clazz)) {
                    _addCollectionChildren();   
                } else if (Map.class.isAssignableFrom(clazz)) {
                    _addMapChildren();
                }                        
        
                // Add the bean properties.
                _addProperties(clazz);
            }                            
        } finally {        
            _initialized = true;
        }        
    }
    
    protected void _initializeFromArray() {
       _allowsChildren = true;
       _isLeaf = false;
                     
       if (_obj instanceof boolean[]) {
            boolean[] array = (boolean[]) _obj;
            int length = Math.min(array.length, getMaxLength());
            _children.ensureCapacity(length);
            for (int i = 0; i < length; i++) {
                _children.add(
                  new BeanTreeNode(this, "" + i, new Boolean(array[i]), 
                  false));                  
            }        
       } else if (_obj instanceof byte[]) {     
            byte[] array = (byte[]) _obj;
            int length = Math.min(array.length, getMaxLength());
            _children.ensureCapacity(length);
            for (int i = 0; i < length; i++) {
                _children.add(
                  new BeanTreeNode(this, "" + i, new Byte(array[i]), false));                  
            }        
       } else if (_obj instanceof char[]) {            
            char[] array = (char[]) _obj;
            int length = Math.min(array.length, getMaxLength());
            _children.ensureCapacity(length);
            for (int i = 0; i < length; i++) {
                _children.add(
                  new BeanTreeNode(this, "" + i, new Character(array[i]),
                  false));                  
            }
       } else if (_obj instanceof short[]) {            
            short[] array = (short[]) _obj;
            int length = Math.min(array.length, getMaxLength());
            _children.ensureCapacity(length);
            for (int i = 0; i < length; i++) {
                _children.add(
                  new BeanTreeNode(this, "" + i, new Short(array[i]), false));                  
            }
       } else if (_obj instanceof int[]) {            
            int[] array = (int[]) _obj;
            int length = Math.min(array.length, getMaxLength());
            _children.ensureCapacity(length);
            for (int i = 0; i < length; i++) {
                _children.add(
                  new BeanTreeNode(this, "" + i, new Integer(array[i]), 
                  false));                  
            }
       } else if (_obj instanceof long[]) {            
            long[] array = (long[]) _obj;
            int length = Math.min(array.length, getMaxLength());
            _children.ensureCapacity(length);
            for (int i = 0; i < length; i++) {
                _children.add(
                  new BeanTreeNode(this, "" + i, new Long(array[i]), false));                        
            }
       } else if (_obj instanceof float[]) {            
            float[] array = (float[]) _obj;
            int length = Math.min(array.length, getMaxLength());
            _children.ensureCapacity(length);
            for (int i = 0; i < length; i++) {
                _children.add(
                  new BeanTreeNode(this, "" + i, new Float(array[i]), false));                        
            }
       } else if (_obj instanceof double[]) {            
            double[] array = (double[]) _obj;
            int length = Math.min(array.length, getMaxLength());
            _children.ensureCapacity(length);
            for (int i = 0; i < length; i++) {
                _children.add(
                  new BeanTreeNode(this, "" + i, new Double(array[i]), 
                    false));                        
            }
       } else if (_obj instanceof Object[]) {
           Object[] array = (Object[]) _obj;
           int length = Math.min(array.length, getMaxLength());
            _children.ensureCapacity(length);
            for (int i = 0; i < length; i++) {
                _children.add(new BeanTreeNode(this, "" + i, array[i],
                 _shouldIntrospect));                  
            }
       } else {
           _logger.warn("Unknown array type: " + _obj);
       }   
    }
    
    /** Add the elements of the collection to the list of children. */
    protected void _addCollectionChildren() {
        Collection c = (Collection) _obj;
        int length = Math.min(c.size(), getMaxLength());        
        _children.ensureCapacity(length);
        int i = 0;
        for (Iterator it = c.iterator(); it.hasNext() && (i < length); ) {
            _children.add(new BeanTreeNode(this, "" + i, it.next(),
            _shouldIntrospect));
            i++;
        }
    }
    
    /** Add the mappings of the map to the list of children. */
    protected void _addMapChildren() {
        Map map = (Map) _obj;
        int length = Math.min(map.size(), getMaxLength());        
        _children.ensureCapacity(length);
        int i = 0;
        for (Iterator it = map.entrySet().iterator(); it.hasNext() && 
             (i < length); ) {
            _children.add(new BeanTreeNode(this, "" + i, it.next(),
            _shouldIntrospect));
            i++;
        }
    }
    
    
    
    protected void _addProperties(Class clazz) {
        if (!_shouldIntrospect) {
            return;
        }
       
        // Keep track of property names so we don't add the same one twice.
        Set set = new HashSet();
        _addBeanProperties(clazz, set);
        _addSpecialProperties(clazz, set);           
        _addFields(clazz, set);
    }    
    
    /** Add the bean properties of the object if they aren't in
     *  <code>propertyNameSet</code>. Add the names of the properties to
     *  <code>propertyNameSet</code>.
     */
    protected void _addBeanProperties(Class clazz, Set propertyNameSet) {
        BeanInfo info = null;
        try {
            info = Introspector.getBeanInfo(clazz);
        } catch (IntrospectionException ie) {
            _logger.warn("Can't introspect " + _obj, ie);
            return;
        }        
        
        // Set the icon if available.
        Image image = info.getIcon(BeanInfo.ICON_COLOR_16x16);
       
        if (image != null) {
            _icon = new ImageIcon(image);
        }
 
        PropertyDescriptor[] properties = info.getPropertyDescriptors();

        _children.ensureCapacity(_children.size() + properties.length);        
                                                
        GetPropertiesLoop:
        for (int i = 0; i < properties.length; i++) {
             PropertyDescriptor property = properties[i];
            
             String propertyName = property.getDisplayName();
             
             if (propertyNameSet.contains(propertyName) ||
                 propertyName.equals("bytes") ||
                 propertyName.startsWith("parent")) {
                 continue GetPropertiesLoop;
             }
            
             Class propertyType = property.getPropertyType();
                         
             Method readMethod = property.getReadMethod();                          
             
             if (readMethod == null) {
                 if (_logger.isInfoEnabled()) {
                     _logger.info("Can't read " + propertyName);
                 }
                 continue GetPropertiesLoop;    
             }             
             
             Object value = null;
             try {
                 value = readMethod.invoke(_obj, null);                                  
             } catch (Exception e) {
                if (_logger.isInfoEnabled()) { 
                    _logger.info("Can't invoke read method for " + 
                      propertyName, e);
                }
                continue GetPropertiesLoop;
             }                                                       
             
             _children.add(new BeanTreeNode(this, propertyName, value, true));
             propertyNameSet.add(propertyName);                       
        }
    }
        
    /** Add properties that require special knowledge if they aren't in
     *  <code>propertyNameSet</code>. Add the names of the properties to
     *  <code>propertyNameSet</code>.
     */
    protected void _addSpecialProperties(Class clazz, Set propertyNameSet) {        
        _addProperty(clazz, "size", propertyNameSet);
        _addProperty(clazz, "length", propertyNameSet);
        _addProperty(clazz, "keys", propertyNameSet);
    }
    
    /** If a method with the argument name and no arguments exists, invoke it
     *  to get a return value. Add a tree node based on the return value to the
     *  children. If an error occurs, do nothing.
     */
    protected void _addProperty(Class clazz, String name, 
      Set propertyNameSet) {
        if (propertyNameSet.contains(name)) {
            return;
        }
        
        Object value = null;
        try {
            Method method = clazz.getMethod(name, null);
            value = method.invoke(_obj, null);
        } catch (Exception e) {
            return;
        }
        
        _children.add(new BeanTreeNode(this, name, value, true));
        propertyNameSet.add(name);
    }
    
    /** Add public fields of clazz if they aren't in propertyNameSet. Add the
     *  names of the fields to <code>propertyNameSet</code>.
     */    
    public void _addFields(Class clazz, Set propertyNameSet) {        
        Field[] fields = clazz.getFields();
        
        FieldLoop:
        for (int i = 0; i < fields.length; i++) {
            Field f = fields[i];
            // Get only instance fields.
            if ((f.getModifiers() & Modifier.STATIC) == 0) {
                String name = f.getName();
                
                if (propertyNameSet.contains(name)) {
                    continue FieldLoop;
                }
                
                try {
                    Object value = f.get(_obj);
                    _children.add(new BeanTreeNode(this, name, value, true));
                    propertyNameSet.add(name);
                } catch (Exception e) {
                    if (_logger.isInfoEnabled()) {
                        _logger.info("Can't get field " + name + 
                        "from " + _obj, e);
                    }
                }
            }          
        }
    }
    
    /** If this node was created by a parent, the property name of this node
     *  according to the parent. Otherwise, whatever name was passed to the
     *  constructor of a root node.
     */
    protected final String _name;
    protected final TreeNode _parent;

    /** The object that this node represents. */    
    protected final Object _obj;
    
    /** An icon that represents this object, if the bean introspector provided
     *  one. Otherwise, <code>null</code>.
     */
    protected Icon _icon = null;

    /** The child nodes. */    
    protected final ArrayList _children = new ArrayList();
    
    /** True if we have finished inspecting the object. */
    protected boolean _initialized = false;
        
    protected boolean _isLeaf = true;
    protected boolean _allowsChildren = false;
    
    /** True if this node should be introspected. */
    protected boolean _shouldIntrospect = true;    
    
    private static final Logger _logger =
      Logger.getLogger(BeanTreeNode.class);
}
